//============================================================================
// 635_WinTest Code.
// Copyright 2001-2004, Crystalfontz America, Inc. Written by Brent A. Crosby
//
// You may use this code in any project (including commercial) as long as
// at least one of the intents of the project is to communicate with a
// Crystalfontz display.
//
// http://www.crystalfontz.com brent@crystalfontz.com
//
// !!!!!!!! THIS IS TEST CODE ONLY !!!!!!!!
// !!!!!!!!  USE AT YOUR OWN RISK  !!!!!!!!
//
//============================================================================
// Packet_Debug.cpp : implementation file
//

#include "stdafx.h"
#include "635_WinTest.h"
#include "serial.h"
#include "GPIO.h"
#include "KeyPads.h"
#include "TempsDlg.h"
#include "Fans.h"
#include "Packet_Debug.h"
#include "Live_Display.h"
#include "LED_Editor.h"
#include "635_WinTestDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
/////////////////////////////////////////////////////////////////////////////
//"Official" (WinTest pull-down & docs) command names
//  0 = Ping Command
//  1 = Get Hardware & Firmware Version
//  2 = Write User Flash Area
//  3 = Read User Flash Area
//  4 = Store Current State As Boot State
//  5 = Reboot
//  6 = Clear LCD Screen
//  7 = Set LCD Contents, Line 1
//  8 = Set LCD Contents, Line 2
//  9 = Set LCD Special Character Data
//  10 = Read 8 Bytes of LCD Memory
//  11 = Set LCD Cursor Position
//  12 = Set LCD Cursor Style
//  13 = Set LCD Contrast
//  14 = Set LCD & Keypad Backlight
//  15 = Query Last Fan Pulse Count
//  16 = Setup Fan Reporting
//  17 = Set Fan Power
//  18 = Read DOW Device Information
//  19 = Setup Temperature Reporting
//  20 = Arbitrary DOW Transaction
//  21 = Setup Live Fan/Temp. Display

// 0x80 = key reporting from the 635
// 0x81 = fan reporting
// 0x82 = temp reporting

// TTcc cccc
// |||| ||||--Command/response/acknowledge or report code 0-63
// ||---------Type:
//              00 = normal command from host to 635
//              01 = normal response/acknowledge from 635 to host
//              10 = report from 635 to host (not in direct
//                   response to a command from the host to the 635)
//              11 = error acknowledge from 635 to host (a packet
//                   with valid structure but illegeal content
//                   was rceived by the 635)
const char
 *command_names[37]=
   {" 0 = Ping",
    " 1 = Read Version",
    " 2 = Write Flash",
    " 3 = Read Flash",
    " 4 = Store Boot State",
    " 5 = Reboot",
    " 6 = Clear LCD",
    " 7 = LCD Line 1",
    " 8 = LCD Line 2",
    " 9 = LCD CGRAM",
    "10 = Read LCD Memory",
    "11 = Place Cursor",
    "12 = Set Cursor Style",
    "13 = Contrast",
    "14 = Backlight",
    "15 = Read Fans",
    "16 = Set Fan Rpt.",
    "17 = Set Fan Power",
    "18 = Read DOW ID",
    "19 = Set Temp. Rpt",
    "20 = DOW Transaction",
    "21 = Set Live Display",
    "22 = Direct LCD Command",
    "23 = Set Key Event Reporting",
    "24 = Read Keypad, Polled Mode",
    "25 = Set Fan Fail-Safe",
    "26 = Set Fan RPM Glitch Filter",
    "27 = Read Fan Pwr & Fail-Safe",
    "28 = Set ATX switch functionality",
    "29 = Watchdog Host Reset",
    "30 = Rd Rpt",
    "31 = Send Data to LCD (row,col)",
    "32 = Key Legends",
    "33 = Set Baud Rate",
    "34 = Set/Configure GPIO",
    "35 = Read GPIO & Configuration",
	"36 = Advanced Settings"};
const char
  *error_names[37]=
   {"Error: Ping",
    "Error: Version",
    "Error: Write Flash",
    "Error: Read Flash",
    "Error: Store Boot State",
    "Error: Reboot",
    "Error: Clear LCD",
    "Error: Set LCD 1",
    "Error: Set LCD 2",
    "Error: Set LCD CGRAM",
    "Error: Read LCD Memory",
    "Error: Place LCD Cursor",
    "Error: LCD Cursor Style",
    "Error: Contrast",
    "Error: Backlight",
    "Error: Query Fan",
    "Error: Set Fan Rept.",
    "Error: Set Fan Power",
    "Error: Read DOW ID",
    "Error: Set Temp. Rept.",
    "Error: DOW Transaction",
    "Error: Setup Live Disp",
    "Error: Direct LCD Command",
    "Error: Set Key Event Reporting",
    "Error: Read Keypad, Polled Mode",
    "Error: Set Fan Fail-Safe",
    "Error: Set Fan RPM Glitch Filter",
    "Error: Read Fan Pwr & Fail-Safe",
    "Error: Set ATX switch functionality",
    "Error: Watchdog Host Reset",
    "Error: Read  Reporting/ATX/Watchdog",
    "Error: Send Data to LCD (row,col)",
    "Error: Key Legends",
    "Error: Set Baud Rate",
    "Error: Set/Configure GPIO",
    "Error: Read GPIO & Configuration",
    "Error: Advanced Settings"};
  const char
    *key_names[21]=
     {"KEY_NONE",
      "KEY_UP_PRESS",
      "KEY_DOWN_PRESS",
      "KEY_LEFT_PRESS",
      "KEY_RIGHT_PRESS",
      "KEY_ENTER_PRESS",
      "KEY_EXIT_PRESS",
      "KEY_UP_RELEASE",
      "KEY_DOWN_RELEASE",
      "KEY_LEFT_RELEASE",
      "KEY_RIGHT_RELEASE",
      "KEY_ENTER_RELEASE",
      "KEY_EXIT_RELEASE",
      "KEY_UL_PRESS",
      "KEY_UR_PRESS",
      "KEY_LL_PRESS",
      "KEY_LR_PRESS",
      "KEY_UL_RELEASE",
      "KEY_UR_RELEASE",
      "KEY_LL_RELEASE",
      "KEY_LR_RELEASE"};
/////////////////////////////////////////////////////////////////////////////
// Packet_Debug dialog
Packet_Debug::Packet_Debug(CWnd* pParent /*=NULL*/)
  : CDialog(Packet_Debug::IDD, pParent)
  {
  //{{AFX_DATA_INIT(Packet_Debug)
    // NOTE: the ClassWizard will add member initialization here
  //}}AFX_DATA_INIT
  ASSERT(pParent != NULL);

  m_pParent = pParent;
  m_nID = Packet_Debug::IDD;
  }

BOOL Packet_Debug::OnInitDialog()
  {
  CDialog::OnInitDialog();
  //Set the font of the log windows to fixed-pitch
  HFONT m_hFont;
  CFont *m_pFont;
  m_hFont = (HFONT)GetStockObject(ANSI_FIXED_FONT);
  m_pFont = CFont::FromHandle (m_hFont);
  m_received.SetFont (m_pFont);
  m_sent.SetFont (m_pFont);
  m_command.SetCurSel(0);

  m_show_fans.SetCheck(1);
  m_show_temperatures.SetCheck(1);
  m_show_acknowledges.SetCheck(1);
  m_show_keys.SetCheck(1);
  m_show_hex_output.SetCheck(0);
  m_show_ascii_output.SetCheck(1);

  acknowledge_spinner_state=0;
  temperature_spinner_state=0;
  fan_spinner_state=0;
  key_spinner_state=0;

  sent_history_head=sent_history_tail=0;
  received_history_head=received_history_tail=0;

  return TRUE;  // return TRUE  unless you set the focus to a control
  } 
void Packet_Debug::DoDataExchange(CDataExchange* pDX)
  {
  CDialog::DoDataExchange(pDX);
  //{{AFX_DATA_MAP(Packet_Debug)
  DDX_Control(pDX, IDC_SHOW_ASCII_OUTPUT, m_show_ascii_output);
  DDX_Control(pDX, IDC_SHOW_HEX_OUTPUT, m_show_hex_output);
  DDX_Control(pDX, IDC_SHOW_KEYS, m_show_keys);
  DDX_Control(pDX, IDC_KEY_SPINNER, m_key_spinner);
  DDX_Control(pDX, IDC_ACKNOWLEDGE_SPINNER, m_acknowledge_spinner);
  DDX_Control(pDX, IDC_FAN_SPINNER, m_fan_spinner);
  DDX_Control(pDX, IDC_TEMP_SPINNER, m_temp_spinner);
  DDX_Control(pDX, IDC_SHOW_TEMPERATURES, m_show_temperatures);
  DDX_Control(pDX, IDC_SHOW_FANS, m_show_fans);
  DDX_Control(pDX, IDC_SHOW_ACKNOWLEDGES, m_show_acknowledges);
  DDX_Control(pDX, IDC_COMMAND_COMBO, m_command);
  DDX_Control(pDX, IDC_INPUT, m_input);
  DDX_Control(pDX, IDD_DATA_LENGTH, m_data_length);
  DDX_Control(pDX, IDC_RECEIVED, m_received);
  DDX_Control(pDX, IDC_SENT, m_sent);
  //}}AFX_DATA_MAP
  }

//Allows both ON_WM_CLOSE and ON_BN_CLICKED(IDOK, ) to map to
//a single function with out breaking the resource editor,
#define OnDone OnClose

BEGIN_MESSAGE_MAP(Packet_Debug, CDialog)
  //{{AFX_MSG_MAP(Packet_Debug)
  ON_BN_CLICKED(IDC_Send_Packet, OnSendPacket)
  ON_EN_CHANGE(IDC_INPUT, OnChangeInput)
  ON_EN_SETFOCUS(IDC_INPUT, OnSetfocusInput)
  ON_EN_UPDATE(IDC_INPUT, OnUpdateInput)
  ON_EN_KILLFOCUS(IDC_INPUT, OnKillfocusInput)
  ON_BN_CLICKED(IDC_SHOW_FANS, OnShowFans)
  ON_BN_CLICKED(IDC_SHOW_TEMPERATURES, OnShowTemperatures)
  ON_BN_CLICKED(IDC_SHOW_ACKNOWLEDGES, OnShowAcknowledges)
  ON_BN_CLICKED(IDC_SHOW_KEYS, OnShowKeys)
  ON_BN_CLICKED(IDC_SHOW_HEX_OUTPUT, OnShowHexOutput)
  ON_BN_CLICKED(IDC_SHOW_ASCII_OUTPUT, OnShowAsciiOutput)
  ON_BN_CLICKED(IDC_CLEAR_HISTORIES, OnClearHistories)
  ON_BN_CLICKED(IDOK, OnDone)
  ON_WM_CLOSE()
  //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
BOOL Packet_Debug::Create()
  {
  return CDialog::Create(m_nID, m_pParent);
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::PostNcDestroy()
  {
  delete this;
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::OnClose() 
  {
  ((CMy635_WinTestDlg*)m_pParent)->Packet_Debugger_Done();
  }
/////////////////////////////////////////////////////////////////////////////
//Converts a user input field to an array:
//
// \\         => \
// \DDD       => decimal byte
// \xHH       => hex byte
// \bBBBBBBBB => binary byte
//
//returns the length of the array
int Packet_Debug::Convert_Input_String_To_Array(const char *input,unsigned char *output)
  {
  int
    i;
  int
    count;
  for(i=0,count=0;input[i]&&(count<=MAX_DATA_LENGTH);i++,count++)
    {
    //detect the escaped backslash as one character
    if((input[i]=='\\')&&(input[i+1]=='\\'))
      {
      output[count]='\\';
      i++;
      }
    else
      {
      //detect the \ddd (decimal) as one character
      if((input[i]=='\\')&&(isdigit(input[i+1]))
                          &&(isdigit(input[i+2]))
                          &&(isdigit(input[i+3]))
                          &&(((input[i+1]-'0')*100
                              +(input[i+2]-'0')*10
                              +(input[i+3]-'0'))<=255))
        {
        output[count]=
          ((input[i+1]-'0')*100+(input[i+2]-'0')*10+(input[i+3]-'0'));
        i+=3;
        }
      else
        {
        //detect the \xXX (hex) as one character
        if((input[i]=='\\')&&(input[i+1]=='x')
                            &&(isxdigit(input[i+2]))
                            &&(isxdigit(input[i+3])))
          {
          unsigned int
            hex_digit;
          char
            hex_input[3];
          hex_input[0]=input[i+2];
          hex_input[1]=input[i+3];
          hex_input[2]=0;
          sscanf(hex_input,"%x",&hex_digit);
          output[count]=hex_digit;
          i+=3;
          }
        else
          {
          //detect the \bBBBBBB (binary) as one character
          if((input[i]=='\\')&&(input[i+1]=='b')
                              &&((input[i+2]=='0')||(input[i+2]=='1'))
                              &&((input[i+3]=='0')||(input[i+3]=='1'))
                              &&((input[i+4]=='0')||(input[i+4]=='1'))
                              &&((input[i+5]=='0')||(input[i+5]=='1'))
                              &&((input[i+6]=='0')||(input[i+6]=='1'))
                              &&((input[i+7]=='0')||(input[i+7]=='1'))
                              &&((input[i+8]=='0')||(input[i+8]=='1'))
                              &&((input[i+9]=='0')||(input[i+9]=='1')))
            {
            output[count]=
              ((input[i+2]=='1')?0x80:0x00)|
              ((input[i+3]=='1')?0x40:0x00)|
              ((input[i+4]=='1')?0x20:0x00)|
              ((input[i+5]=='1')?0x10:0x00)|
              ((input[i+6]=='1')?0x08:0x00)|
              ((input[i+7]=='1')?0x04:0x00)|
              ((input[i+8]=='1')?0x02:0x00)|
              ((input[i+9]=='1')?0x01:0x00);
            i+=9;
            }
          else
            output[count]=input[i];
          }
        }
      }
    }
  return(count);
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::OnChangeInput()
  {
  static char
    hereOnChangeInput=0;
  if(!hereOnChangeInput)
    {
    hereOnChangeInput=1;
    char
      string[1024];
    m_input.GetWindowText(string,1024);

    //Update the prompt
    sprintf(string,
            "Length=%d",
            Convert_Input_String_To_Array(string,(unsigned char *)string));
    m_data_length.SetWindowText(string);
    hereOnChangeInput=0;
    }
  }
//----------------------------------------------------------------------------
void Packet_Debug::OnSetfocusInput() 
  {
  OnChangeInput();  
  }
//----------------------------------------------------------------------------
void Packet_Debug::OnUpdateInput() 
  {
  OnChangeInput();  
  }
//----------------------------------------------------------------------------
void Packet_Debug::OnKillfocusInput() 
  {
  OnChangeInput();  
  }
//============================================================================
void Packet_Debug::OnShowFans() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnShowTemperatures() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnShowAcknowledges() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnShowKeys() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnShowHexOutput() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnShowAsciiOutput() 
  {
  Show_Received_Packets();
  }
//============================================================================
void Packet_Debug::OnClearHistories() 
  {
  sent_history_head=sent_history_tail=0;
  received_history_head=received_history_tail=0;
  Show_Sent_Packets();
  Show_Received_Packets();
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Set_Sent_User_Text(HISTORY_ITEM *history)
  {
  // 0 = waiting to be sent
  // 1 = sent, waiting for acknowledge
  // 2 = timed out  
  // 3 = (command, will not happen)
  // 4 = acknowleged without error
  // 5 = (response, will not happen)
  // 6 = acknowledged with error
  const char
    ack_names[7]={'q','w','t','c','a','r','e'};

  //Make the command sub-string based on hex or decimal
  char
    command_string[8];
  
  if(m_show_hex_output.GetCheck())
    sprintf(command_string,"C=x%02X",history->packet.command);
  else
    sprintf(command_string,"C=%d",history->packet.command);

  char
    expanded[1024];
  expanded[0]=0;
  int
    i;
  for(i=0;i<history->packet.data_length;i++)
    {
    int
      current_length=strlen(expanded);
    if(isprint(history->packet.data[i])&&(m_show_ascii_output.GetCheck()))
      sprintf(&expanded[current_length],"%c",history->packet.data[i]);
    else
      {
      if(m_show_hex_output.GetCheck())
        sprintf(&expanded[current_length],"\\x%02X",history->packet.data[i]);
      else
        sprintf(&expanded[current_length],"\\%03d",history->packet.data[i]);
      }
    }
  sprintf(history->user_text,"%c%s(%s),L=%d,D=\"%s\",CRC=0x%04X\r\n",
          ack_names[history->status],
          command_string,
          command_names[history->packet.command],
          history->packet.data_length,
          expanded,
          history->packet.CRC.as_word);
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Show_Sent_Packets(void) 
  {
  //No sense malloc + free all the time
  static char
    sent_window_contents[1024*SENT_HISTORY_LENGTH];

  sent_window_contents[0]=0;
  char
    *concat_pointer;
  concat_pointer=sent_window_contents;
  int
    entry;

  //Loop over the entire output buffer, adding
  //the lines.
  entry=sent_history_tail;
  while(entry!=sent_history_head)
    {
    strcpy(concat_pointer,sent_history[entry].user_text);
    concat_pointer+=strlen(sent_history[entry].user_text);

    //Move to the next packet
    entry++;
    if(SENT_HISTORY_LENGTH<=entry)
      entry=0;
    }

  //Remove the trailing CR/LF
  if(concat_pointer!=sent_window_contents)
    concat_pointer[-2]=0;

  m_sent.SetWindowText(sent_window_contents);
  //Scroll to the bottom (most recent data)
  m_sent.LineScroll(SENT_HISTORY_LENGTH);
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Send_Queued_Packets(void) 
  {
  //See how many packets are unacknowleged
  int
    entry;
  entry=sent_history_tail;
  int
    Unacknowledged_Packets;
  Unacknowledged_Packets=0;
  while(entry!=sent_history_head)
    {
    if(sent_history[entry].status==
       SENT_STATUS_WAITING_FOR_ACKNOWLEDGE) 
      Unacknowledged_Packets++;
    //Move to the next packet
    entry++;
    if(SENT_HISTORY_LENGTH<=entry)
      entry=0;
    }

  //Send as many as we can up to MAX_UNACKNOWLEDGED_PACKETS.
  entry=sent_history_tail;
  while((Unacknowledged_Packets<MAX_UNACKNOWLEDGED_PACKETS)&&
        (entry!=sent_history_head))
    {
    //See if this packet is "waiting to be sent"
    if(sent_history[entry].status==
       SENT_STATUS_WAITING_TO_BE_SENT) 
      {
      //Drop the packet out the hole
      send_packet(&(sent_history[entry].packet));
      //Mark it as "sent, waiting for acknowledge"
      sent_history[entry].status=
        SENT_STATUS_WAITING_FOR_ACKNOWLEDGE;
      //Give the module some time to respond
      sent_history[entry].timeout=ACKNOWLEDGE_TIMEOUT;
      //Update the user_text for this entry
      Set_Sent_User_Text(&sent_history[entry]);
      //That is one more packet that has been sent to the LCD 
      //module and has not yet received an acknowledge.
      Unacknowledged_Packets++;
      }
    //Move to the next packet
    entry++;
    if(SENT_HISTORY_LENGTH<=entry)
      entry=0;
    }
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Queue_Outgoing_Packet(COMMAND_PACKET *packet) 
  {
  //This is waiting to be sent
  sent_history[sent_history_head].status=SENT_STATUS_WAITING_TO_BE_SENT;
  //Log the packet
  sent_history[sent_history_head].packet=*packet;
  //Set the CRC to 0 (it gets displayed while the packet
  //is queued).
  sent_history[sent_history_head].packet.CRC.as_word=0;
  //Update the user_text for this entry
  Set_Sent_User_Text(&sent_history[sent_history_head]);

  //tick the head index around the buffer
  sent_history_head++;
  if(SENT_HISTORY_LENGTH<=sent_history_head)
    sent_history_head=0;

  //let the head index push the tail index around too
  if(sent_history_head==sent_history_tail)
    {
    sent_history_tail++;
    if(SENT_HISTORY_LENGTH<=sent_history_tail)
      sent_history_tail=0;
    }

  //Attempt to immediately send the packet.
  Send_Queued_Packets();

  Show_Sent_Packets();
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Timeout_No_Repsonse_Packets(void)
  {
  //We need to scan forwards through the sent history and
  //count down the timers of any packets that are
  //"sent, waiting for acknowledge", and expire them if they
  //are too old.
  int
    need_to_update=0;
  int
    entry;
  entry=sent_history_tail;
  while(entry!=sent_history_head)
    {
    //See if this packet is "sent, waiting for acknowledge"
    if(sent_history[entry].status==
       SENT_STATUS_WAITING_FOR_ACKNOWLEDGE) 
      {
      //Count its timer down
      if(sent_history[entry].timeout)
        {
        sent_history[entry].timeout--;
        if(!sent_history[entry].timeout)
          {
          //Mark it as dead
          sent_history[entry].status=
            SENT_STATUS_TIMED_OUT; 
          //Refresh the text for this entry
          Set_Sent_User_Text(&sent_history[entry]);
          need_to_update=1;
          }
		  Sleep(1);
        }
      }
    //Move to the next packet
    entry++;
    if(SENT_HISTORY_LENGTH<=entry)
      entry=0;
    }
  //If we timed out a packet, then we can send the next one that
  //is waiting, and then update the display.
  if(need_to_update)
    {
    Send_Queued_Packets();
    Show_Sent_Packets();
    }
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Reconcile_Acknowledge(COMMAND_PACKET *packet) 
  {
  //We have just received an acknowledge or error acknowlege.
  //
  // TTcc cccc
  // |||| ||||--Command/response/acknowledge or report code 0-63
  // ||---------Type:
  //              00 = normal command from host to 635
  //              01 = normal response/acknowledge from 635 to host
  //              10 = report from 635 to host (not in direct
  //                   response to a command from the host to the 635)
  //              11 = error acknowledge from 635 to host (a packet
  //                   with valid structure but illegeal content
  //                   was rceived by the 635)
  //



  //We need to scan forwards through the sent history and
  //the first matching packet that is "sent, waiting for
  //acknowledge" with the response from the LCD module.
  //
  //Any packets before that point need to be marked as timed
  //out.
  int
    entry;
  entry=sent_history_tail;
  while(entry!=sent_history_head)
    {
    //See if this packet is "sent, waiting for acknowledge"
    if(sent_history[entry].status==
       SENT_STATUS_WAITING_FOR_ACKNOWLEDGE) 
      {
      //the command has to match up
      if(((sent_history[entry].packet.command)&0x3F)==
         ((packet->command)&0x3F))
        {
        //Now set the acknowledge
        sent_history[entry].status=((packet->command)>>6)+3;
        //Refresh the text for this this_entry
        Set_Sent_User_Text(&sent_history[entry]);
        //Stop looking
        break;
        }
      }
    //Move to the next packet
    entry++;
    if(SENT_HISTORY_LENGTH<=entry)
      entry=0;
    }
  //See if there is a packet waiting to be sent
  Send_Queued_Packets();
  //Refresh the window
  Show_Sent_Packets();
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Set_Received_User_Text(char *user_text,COMMAND_PACKET *packet)
  {
  //First expand the data field into \ddd format 
  char
    expanded[MAX_DATA_LENGTH*4+1];  // each char tp \xxx + terminator
  int
    i;
  expanded[0]=0;
  for(i=0;i<packet->data_length;i++)
    {
    int
      current_length=strlen(expanded);
    if(isprint(packet->data[i])&&(m_show_ascii_output.GetCheck()))
      sprintf(&expanded[current_length],"%c",packet->data[i]);
    else
      {
      if(m_show_hex_output.GetCheck())
        sprintf(&expanded[current_length],"\\x%02X",packet->data[i]);
      else
        sprintf(&expanded[current_length],"\\%03d",packet->data[i]);
      }
    }

  //Make the command sub-string based on hex or decimal
  char
    command_string[8];
  
  if(m_show_hex_output.GetCheck())
    sprintf(command_string,"C=x%02X",packet->command);
  else
    sprintf(command_string,"C=%d",packet->command);

  //Now do special case formatting for readability
  switch(packet->command)
    {
    case 0x80: //keys
      sprintf(user_text,"%s(key:%s),L=%d,D=\"%s\",CRC=0x%04X\r\n",
              command_string,
              packet->data[0]<=20?
                key_names[packet->data[0]]:
                "HUH?",
              packet->data_length,
              expanded,
              packet->CRC.as_word);
      break;
    case 0x81: //fan reporting
      char
        fan_string[20];
      (((CMy635_WinTestDlg*)m_pParent)->m_pFansDlg)->get_fan(packet, fan_string);
      sprintf(user_text,"%s:f=%d(f=%s) t=%d,n=%d\r\n",
              command_string,
              packet->data[0],
              fan_string,
              packet->data[1],
              (*(WORD *)(&(packet->data[2]))));
      break;
    case 0x40|18: //DOW addresses
      sprintf(user_text,"%s(d=%d):%02X,%02X,%02X,%02X;%02X,%02X,%02X,%02X\r\n",
              command_string,
              packet->data[0],
              packet->data[1],
              packet->data[2],
              packet->data[3],
              packet->data[4],
              packet->data[5],
              packet->data[6],
              packet->data[7],
              packet->data[8]);
      break;
    case 0x40|20: //DOW transaction
      sprintf(user_text,"%s(d=%d):",
              command_string,
              packet->data[0]);
      for(i=1;i<packet->data_length;i++)
        sprintf(&(user_text[strlen(user_text)]),"%s%02X",
          i==1?"":",",
          packet->data[i]);
        sprintf(&(user_text[strlen(user_text)]),"\r\n");
      break;
    case 0x82: //Temp reporting
      double
        degc;
      degc=(*(short *)&packet->data[1])/16.0;

      double
        degf;
      degf=(degc*9.0)/5.0+32.0;

      char
        crc[20];
      if(packet->data[3]==0)
        strcpy(crc,"BAD!");
      else if(packet->data[3]==1)
        strcpy(crc,"good");
      else if(packet->data[3]==2)
        strcpy(crc,"none");
      else
        strcpy(crc,"Huh?");

      sprintf(user_text,"%s(d=%2d):[%04X]=%9.4fC =%9.4fF DOW CRC=%s\r\n",
              command_string,
              packet->data[0],
              *(short *)&packet->data[1],
              degc,
              degf,
              crc);
      break;
    default:
      sprintf(user_text,"%s(%s),L=%d,D=\"%s\",CRC=0x%04X\r\n",
              command_string,
              ((packet->command&0xC0)==0xC0)?
                error_names[0x3F&packet->command]:
                command_names[0x3F&packet->command],
              packet->data_length,
              expanded,
              packet->CRC.as_word);
    }
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Show_Received_Packets(void)
  {
  //No sense malloc + free all the time
  static char
    received_window_contents[1024*RECEIVED_HISTORY_LENGTH];

  //Get local copies of the acknowledge states on the stack. This
  //code has a deep loop, no sense making function calls for these
  //all the time.
  int
    show_fans;
  int
    show_temperatures;
  int
    show_acknowledges;
  int
    show_keys;

  show_fans=m_show_fans.GetCheck();
  show_temperatures=m_show_temperatures.GetCheck();
  show_acknowledges=m_show_acknowledges.GetCheck();
  show_keys=m_show_keys.GetCheck();

  received_window_contents[0]=0;
  char
    *concat_pointer;
  concat_pointer=received_window_contents;

  int
    entry;

  //Loop over the entire output buffer, adding
  //the lines if their type is checked for showing.
  entry=received_history_tail;
  while(entry!=received_history_head)
    {
    switch(received_history[entry].packet.command)
      {
      case 0x80: //keys
        if(show_keys)
          {
          strcpy(concat_pointer,received_history[entry].user_text);
          concat_pointer+=strlen(received_history[entry].user_text);
          }
        break;
      case 0x81: //fan reporting
        if(show_fans)
          {
          strcpy(concat_pointer,received_history[entry].user_text);
          concat_pointer+=strlen(received_history[entry].user_text);
          }
        break;
      case 0x82: //Temp reporting
        if(show_temperatures)
          {
          strcpy(concat_pointer,received_history[entry].user_text);
          concat_pointer+=strlen(received_history[entry].user_text);
          }
        break;
      default:
        if(show_acknowledges)
          {
          strcpy(concat_pointer,received_history[entry].user_text);
          concat_pointer+=strlen(received_history[entry].user_text);
          }
        break;
      }

    //Move to the next packet
    entry++;
    if(RECEIVED_HISTORY_LENGTH<=entry)
      entry=0;
    }

  //Remove the trailing CR/LF
  if(concat_pointer!=received_window_contents)
    concat_pointer[-2]=0;

  //Compare what we have with the previous, and only update the
  //window if it is different. This allows text selection while packets
  //are coming in but masked.
  //No sense malloc + free all the time
  static char
    old_received_window_contents[1024*RECEIVED_HISTORY_LENGTH];
  int
    old_received_window_contents_length;
  old_received_window_contents_length=m_received.GetWindowTextLength();
  m_received.GetWindowText(old_received_window_contents,old_received_window_contents_length+2);
  if(strcmp(old_received_window_contents,received_window_contents))
    {
    m_received.SetWindowText(received_window_contents);
    //Scroll to the bottom (most recent data)
    m_received.LineScroll(RECEIVED_HISTORY_LENGTH);
    }
  }
/////////////////////////////////////////////////////////////////////////////
void Packet_Debug::Record_Received_Packet(COMMAND_PACKET *packet) 
  {
  //Log the packet
  received_history[received_history_head].packet=*packet;
  //Update the user_text for this entry
  Set_Received_User_Text(received_history[received_history_head].user_text,
                         &received_history[received_history_head].packet);

  //tick the head index around the buffer
  received_history_head++;
  if(RECEIVED_HISTORY_LENGTH<=received_history_head)
    received_history_head=0;

  //let the head index push the tail index around too
  if(received_history_head==received_history_tail)
    {
    received_history_tail++;
    if(RECEIVED_HISTORY_LENGTH<=received_history_tail)
      received_history_tail=0;
    }
  Show_Received_Packets();

  //If this packet is an acknowledge, mark its command as acknowledged.
  // TTcc cccc
  // |||| ||||--Command/response/acknowledge or report code 0-63
  // ||---------Type:
  //              00 = normal command from host to 635
  //              01 = normal response/acknowledge from 635 to host
  //              10 = report from 635 to host (not in direct
  //                   response to a command from the host to the 635)
  //              11 = error acknowledge from 635 to host (a packet
  //                   with valid structure but illegeal content
  //                   was rceived by the 635)
  if(((packet->command&0x40)==0x40)||
     ((packet->command&0xC0)==0xC0))
    Reconcile_Acknowledge(packet);
  }
/////////////////////////////////////////////////////////////////////////////
//If the debug window is up, always spin the spinners
//even if the packet is filtered
static const char
  *spinner[4]={"\\","|","/","--"};
//---------------------------------------------------------------------------
void Packet_Debug::RollAcknowledgeSpinner(void) 
  {
  //roll the spinner
  acknowledge_spinner_state++;
  if(3<acknowledge_spinner_state)
    acknowledge_spinner_state=0;
  m_acknowledge_spinner.SetWindowText(spinner[acknowledge_spinner_state]);
  }
//---------------------------------------------------------------------------
void Packet_Debug::RollTemperatureSpinner(void) 
  {
  //roll the spinner
  temperature_spinner_state++;
  if(3<temperature_spinner_state)
    temperature_spinner_state=0;
  m_temp_spinner.SetWindowText(spinner[temperature_spinner_state]);
  }
//---------------------------------------------------------------------------
void Packet_Debug::RollFanSpinner(void) 
  {
  //roll the spinner
  fan_spinner_state++;
  if(3<fan_spinner_state)
    fan_spinner_state=0;
  m_fan_spinner.SetWindowText(spinner[fan_spinner_state]);
  }
//---------------------------------------------------------------------------
void Packet_Debug::RollKeySpinner(void) 
  {
  //roll the spinner
  key_spinner_state++;
  if(3<key_spinner_state)
    key_spinner_state=0;
  m_key_spinner.SetWindowText(spinner[key_spinner_state]);
  }
/////////////////////////////////////////////////////////////////////////////
//Assemble the command & data, then pass to the parent window.
void Packet_Debug::OnSendPacket() 
  {
  COMMAND_PACKET
    outgoing_packet;
  char
    string[1024];

  outgoing_packet.command=m_command.GetCurSel();
  m_input.GetWindowText(string,1024);
  outgoing_packet.data_length=
    Convert_Input_String_To_Array(string,outgoing_packet.data);
  ((CMy635_WinTestDlg*)m_pParent)->QueuePacket(&outgoing_packet);
  }
/////////////////////////////////////////////////////////////////////////////
